/*
// A simple C++-style interface for polynomials
//
//
//
*/

#include "poly.h"
#include <cmath>

// use this command to avoiud std::cout
using namespace std ;

/*                      _                   _             
//   ___ ___  _ __  ___| |_ _ __ _   _  ___| |_ ___  _ __ 
//  / __/ _ \| '_ \/ __| __| '__| | | |/ __| __/ _ \| '__|
// | (_| (_) | | | \__ \ |_| |  | |_| | (__| || (_) | |   
//  \___\___/|_| |_|___/\__|_|   \__,_|\___|\__\___/|_|   
*/

Poly::Poly() {
  order  = -1 ;
  coeffs = NULL ;
}

Poly::Poly(Poly const & a) {
  order = a.getOrder() ;
  coeffs = new valueType[order+1] ;
  for ( indexType i = 0 ; i <= order ; ++i )
    coeffs[i] = a.coeffs[i] ;
}

Poly::Poly(indexType order) {
  this -> order  = order ;
  this -> coeffs = new valueType[order+1] ; // allocation
  for ( indexType i = 0 ; i <= order ; ++i ) coeffs[i] = 0 ;
}

Poly::Poly(valueType coeffs[], indexType numCoeffs) {
  order  = numCoeffs-1 ;
  // this is a pointer to the actual class.
  // equivalence (*this).coeffs == this -> coeffs  ;
  this -> coeffs = new valueType[numCoeffs] ; // allocation

  // copy values
  for ( indexType i=0; i < numCoeffs ; ++i )
    this -> coeffs[i] = coeffs[i] ;
}

/*       _           _                   _             
//    __| | ___  ___| |_ _ __ _   _  ___| |_ ___  _ __ 
//   / _` |/ _ \/ __| __| '__| | | |/ __| __/ _ \| '__|
//  | (_| |  __/\__ \ |_| |  | |_| | (__| || (_) | |   
//   \__,_|\___||___/\__|_|   \__,_|\___|\__\___/|_|   
*/

Poly::~Poly() {
  if ( coeffs != NULL ) delete [] coeffs ;
}

/*            _               
//   ___  ___| |_ _   _ _ __  
//  / __|/ _ \ __| | | | '_ \ 
//  \__ \  __/ |_| |_| | |_) |
//  |___/\___|\__|\__,_| .__/ 
//                     |_|    
*/

void
Poly::setup (valueType cfs[], indexType numCoeffs) {
  // this is a pointer to the actual class.
  // equivalence (*this).coeffs == this -> coeffs  ;
  if ( order+1 < numCoeffs ) {
    if ( coeffs != NULL ) delete [] coeffs ;
    coeffs = new valueType[numCoeffs] ; // allocation
    order  = numCoeffs-1 ;
  }

  // copy values
  indexType i = 0 ;
  for ( ; i < numCoeffs ; ++i ) coeffs[i] = cfs[i] ;
  for ( ; i <= order    ; ++i ) coeffs[i] = 0 ;
}

/*                 _         
//   _ __ ___  ___(_)_______ 
//  | '__/ _ \/ __| |_  / _ \
//  | | |  __/\__ \ |/ /  __/
//  |_|  \___||___/_/___\___|
*/

void
Poly::resize( indexType newOrder ) {
  if ( newOrder > order ) {
    valueType * bf = coeffs ;
    coeffs = new valueType[newOrder+1] ; // allocate memory
    indexType i = 0 ;
    for ( ; i <= order    ; ++i ) coeffs[i] = bf[i] ;
    for ( ; i <= newOrder ; ++i ) coeffs[i] = 0 ;
    delete [] bf ;
  }
  order = newOrder ;
}

/*       _                 
//    __| |_ __ ___  _ __  
//   / _` | '__/ _ \| '_ \ 
//  | (_| | | | (_) | |_) |
//   \__,_|_|  \___/| .__/ 
//                  |_|    
*/
// set to zero all element less thn epsi
void
Poly::drop( valueType epsi ) {
  for ( indexType i = 0 ; i <= order ; ++i )
    if ( abs(coeffs[i]) <= epsi )
      coeffs[i] = 0 ;
}

/*              _       _   
//   _ __  _ __(_)_ __ | |_ 
//  | '_ \| '__| | '_ \| __|
//  | |_) | |  | | | | | |_ 
//  | .__/|_|  |_|_| |_|\__|
//  |_|                     
*/

void
Poly::print( ostream & s ) const {
  // print in a human readable form the polinomial p to the file fd
  // +1 + -1 x + -2 * x**2 + 0 * x**3
  for ( indexType i = 0 ; i <= order ; ++i ) {
    valueType v = coeffs[i] ;
    switch (i) {
      case 0:  s << v ; break ;
      case 1:
        // skip zero values, eliminates +-1
        if ( v > 0 ) {
          if ( v == 1 )  s << " + x" ;
          else           s << " + " << v << "*x" ;
        } else if ( v < 0 ) {
          if ( v == -1 ) s << " - x" ;
          else           s << " - " << -v << "*x" ;
        }
      break ;
      default:
        // skip zero values, eliminates +-1
        if ( v > 0 ) {
          if ( v == 1 )  s << " + x**" << i ;
          else           s << " + " << v << "*x**" << i ;
        } else if ( v < 0 ) {
          if ( v == -1 ) s << " - x**" << i ;
          else           s << " - " << -v << "*x**" << i ;
        }
      break ;
    }
  }
  if ( order < 0 ) s << '0' ;
}

/*              _    ___          _           
//    __ _  ___| |_ / _ \ _ __ __| | ___ _ __ 
//   / _` |/ _ \ __| | | | '__/ _` |/ _ \ '__|
//  | (_| |  __/ |_| |_| | | | (_| |  __/ |   
//   \__, |\___|\__|\___/|_|  \__,_|\___|_|   
//   |___/                                    
*/
// return the true order of the polynomial
indexType
Poly::getOrder() const {
  indexType trueOrder = order ;
  for ( ; trueOrder >= 0 ; --trueOrder )
    if ( coeffs[trueOrder] != 0 )
      break ;
  return trueOrder ;
}

/*
//   _                _ _             
//  | | ___  __ _  __| (_)_ __   __ _ 
//  | |/ _ \/ _` |/ _` | | '_ \ / _` |
//  | |  __/ (_| | (_| | | | | | (_| |
//  |_|\___|\__,_|\__,_|_|_| |_|\__, |
//                              |___/
*/
// return the leading coefficient of the polynomial
valueType
Poly::leading() const {
  return coeffs[getOrder()] ;
}

/*                              _                     
//    ___  _ __   ___ _ __ __ _| |_ ___  _ __   _____ 
//   / _ \| '_ \ / _ \ '__/ _` | __/ _ \| '__| |_____|
//  | (_) | |_) |  __/ | | (_| | || (_) | |     _____
//   \___/| .__/ \___|_|  \__,_|\__\___/|_|    |_____|  
//        |_|  
*/

Poly const &
Poly::operator = ( Poly const & p ) {

  // if necessary reallocate memory
  indexType pOrder = p.getOrder() ;
  if ( order < pOrder ) {
    if ( coeffs != NULL ) delete [] coeffs ;
    coeffs = new valueType[pOrder+1] ;
    order  = pOrder;
  }

  indexType i = 0 ;
  for ( ; i <= pOrder ; ++i ) coeffs[i] = p.coeffs[i] ;
  for ( ; i <= order  ; ++i ) coeffs[i] = 0 ;
  
  return *this ; // return reference to itself
                 // this permits multiple assignment
                 // like p = q = r ;
}

Poly const &
Poly::operator = ( valueType s ) {
  for ( indexType i = 0 ; i <= order ; ++i ) coeffs[i] = 0 ;
  coeffs[0] = s ;
  return *this ;
}

/*                              _
//    ___  _ __   ___ _ __ __ _| |_ ___  _ __     _   _____ 
//   / _ \| '_ \ / _ \ '__/ _` | __/ _ \| '__|  _| |_|_____|
//  | (_) | |_) |  __/ | | (_| | || (_) | |    |_   _|_____
//   \___/| .__/ \___|_|  \__,_|\__\___/|_|      |_| |_____|     
//        |_|                                               
*/

Poly const &
Poly::operator += ( Poly const & p ) {
  indexType pOrder = p.getOrder() ;
  if ( pOrder > order ) resize( pOrder ) ;
  for ( indexType i = 0 ; i <= pOrder ; ++i )
    coeffs[i] += p.coeffs[i] ;
  return *this ;
}

/*                              _
//    ___  _ __   ___ _ __ __ _| |_ ___  _ __         _____ 
//   / _ \| '_ \ / _ \ '__/ _` | __/ _ \| '__|  _____|_____|
//  | (_) | |_) |  __/ | | (_| | || (_) | |    |_____|_____
//   \___/| .__/ \___|_|  \__,_|\__\___/|_|          |_____|     
//        |_|                                               
*/

Poly const &
Poly::operator -= ( Poly const & p ) {
  indexType pOrder = p.getOrder() ;
  if ( pOrder > order ) resize( pOrder ) ;
  for ( indexType i = 0 ; i <= pOrder ; ++i )
    coeffs[i] -= p.coeffs[i] ;
  return *this ;
}

/*
//                              _                ____  
//    ___  _ __   ___ _ __ __ _| |_ ___  _ __   / /\ \ 
//   / _ \| '_ \ / _ \ '__/ _` | __/ _ \| '__| | |  | |
//  | (_) | |_) |  __/ | | (_| | || (_) | |    | |  | |
//   \___/| .__/ \___|_|  \__,_|\__\___/|_|    | |  | |
//        |_|                                   \_\/_/ 
*/

valueType
Poly::operator () ( valueType x ) const {
  // compute p(v) using Horner rule
  // p(x) = p0 + p1 * x + p2 * x**2 + p3 * x**3 ;
  // p(x) = p0 + x*(p1 + x * (p2 + x * p3)) ;
  valueType res = coeffs[order] ;
  for ( indexType i=order-1 ; i >= 0 ; --i )
    res = res * x + coeffs[i] ;
  return res ;
}

/*
           _                        _ 
  _____  _| |_ ___ _ __ _ __   __ _| |
 / _ \ \/ / __/ _ \ '__| '_ \ / _` | |
|  __/>  <| ||  __/ |  | | | | (_| | |
 \___/_/\_\\__\___|_|  |_| |_|\__,_|_|
                            _                 
  ___  _ __   ___ _ __ __ _| |_ ___  _ __ ___ 
 / _ \| '_ \ / _ \ '__/ _` | __/ _ \| '__/ __|
| (_) | |_) |  __/ | | (_| | || (_) | |  \__ \
 \___/| .__/ \___|_|  \__,_|\__\___/|_|  |___/
      |_|                                     
*/

// negate a polynomial
Poly
operator - ( Poly const & a ) {
  indexType aOrder   = a.getOrder() ;
  Poly res(aOrder) ;
  for ( indexType i = 0 ; i <= aOrder ; ++i ) res[i] = -a[i] ;
  return res ;
}

Poly
operator + ( Poly const & a, Poly const & b ) {
  indexType aOrder   = a.getOrder() ;
  indexType bOrder   = b.getOrder() ;
  indexType resOrder = max(aOrder,bOrder) ;
  Poly res(resOrder) ;
  for ( indexType i = 0 ; i <= resOrder ; ++i ) res[i] = a[i] + b[i] ;
  return res ;
}

Poly
operator - ( Poly const & a, Poly const & b ) {
  indexType aOrder   = a.getOrder() ;
  indexType bOrder   = b.getOrder() ;
  indexType resOrder = max(aOrder,bOrder) ;
  Poly res(resOrder) ;
  for ( indexType i = 0 ; i <= resOrder ; ++i )
    res[i] = a[i] - b[i] ;
  return res ;
}

// multiply two polynomial
// p(x) = a(x) * b(x)
// a(x) = a0 + a1 * x + a2 * x**2
// b(x) = b0 + b1 * x + b2 * x**2
// a(x)*b(x) = p0 + p1 * x + p2 * x**2 + ... + p4 * x**4
// p0 = a0 * b0
// p1 = a0 * b1 + a1 * b0
// p2 = a0 * b2 + a1 * b1 + a2 * b0
//
// pk = sum( i+j = k, i>=0, j>=0 ) ai * bj
Poly
operator * ( Poly const & a, Poly const & b ) {
  indexType aOrder   = a.getOrder() ;
  indexType bOrder   = b.getOrder() ;
  indexType resOrder = aOrder+bOrder ;
  Poly res(resOrder) ;
  for ( indexType i = 0 ; i <= resOrder ; ++i )
    res[i] = a[i] - b[i] ;

  for ( indexType i = 0 ; i <= resOrder ; ++i ) {
    valueType v = 0 ;
    for ( indexType k = max(0,i-bOrder) ; k <= min(i,aOrder) ; ++k )
      v += a[k] * b[i-k] ;
    res[i] = v ;
  }

  return res ;
}

Poly
operator + ( valueType a, Poly const & b ) {
  Poly res(b) ;
  res[0] += a ;
  return res ;
}

Poly
operator - ( valueType a, Poly const & b ) {
  Poly res(b) ;
  res[0] = a - res[0] ;
  return res ;
}

Poly
operator * ( valueType a, Poly const & b ) {
  Poly res(b) ;
  indexType bOrder = b.getOrder() ;
  for ( indexType i=0 ; i <= bOrder ; ++i ) res[i] *= a ;
  return res ;
}

Poly
operator + ( Poly const & a, valueType b ) {
  Poly res(a) ;
  res[0] += b ;
  return res ;
}

Poly
operator - ( Poly const & a, valueType b ) {
  Poly res(a) ;
  res[0] -= b ;
  return res ;
}

Poly
operator * ( Poly const & a, valueType b ) {
  Poly res(a) ;
  indexType aOrder = a.getOrder() ;
  for ( indexType i=0 ; i <= aOrder ; ++i ) res[i] *= b ;
  return res ;
}

Poly
operator / ( Poly const & a, valueType b ) {
  Poly res(a) ;
  indexType aOrder = a.getOrder() ;
  for ( indexType i=0 ; i <= aOrder ; ++i ) res[i] /= b ;
  return res ;
}

/*                  _          __               _                 
//    ___ _ __   __| |   ___  / _|  _ __   ___ | |_   _   ___ ___ 
//   / _ \ '_ \ / _` |  / _ \| |_  | '_ \ / _ \| | | | | / __/ __|
//  |  __/ | | | (_| | | (_) |  _| | |_) | (_) | | |_| || (_| (__ 
//   \___|_| |_|\__,_|  \___/|_|   | .__/ \___/|_|\__, (_)___\___|
//                                 |_|            |___/           
*/
